/* This file includes ethernet task and its related functions. Please refer to 
 * the header file for description of functions. 
 */
 
/* Standard includes. */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* Scheduler includes. */
#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"

/* Demo includes. */
#include "SAM7_EMAC.h"

/* lwIP includes. */
#include "lwip/api.h" 
#include "lwip/tcpip.h"
#include "lwip/memp.h" 
#include "lwip/stats.h"
#include "netif/loopif.h"
#include "pio.h"
#include "mac_tasks.h"
#include "adc.h"
#include "spi.h"
#include "ADS1258.h"
#include "EMflash.h"


/* comm_open is used to hold the state of connection. When we are connected it
 * becomes 1. Then it is used to control the state of connection. If it is less
 * than 1 we don't send any data from ethernet. 
 */
static unsigned char comm_open;

/* config_mode is used to control the configuration mode of the device
 * from ethernet connection.
 */
static char config_mode = 0;

void vlwIPInit( void )
{
    /* Initialize lwIP and its interface layer. */
	sys_init();
	mem_init();								
	memp_init();
	pbuf_init(); 
	netif_init();
	ip_init();
	tcpip_init( NULL, NULL );
}
void adc_monitor_via_UDP( void *pvParameters )
{		
	/* MAC related variables */
	struct ip_addr my_ip_addr, my_net_mask, my_gateway;
	extern err_t ethernetif_init( struct netif *netif );
	static struct netif emac_if;
	/* UDP transfer variables */
	struct netconn *listen_pc;
	struct netbuf *buf;
	struct ip_addr addr;
			
	/* Parameters are not used - suppress compiler error. */
	( void ) pvParameters;	
 			  		
	/* Create a binary semaphore to use common data buffer with spi 
	 * This semaphore is taken from ISR when the buffer is filled 
	 */
	vSemaphoreCreateBinary( spi_buffer_semaphore );
	if( spi_buffer_semaphore == NULL ) { 
		while( true )
			pioA_set_led( 0 , 0 ); 
	} 	
	/* No communication avaliable in the beginning */
	comm_open = 0 ;
		
	/* Create and configure the EMAC interface. */
	
	IP4_ADDR(&my_ip_addr,(SETTINGS_PAGE[LOCAL_IP] & 0xFF000000) >> 24 \
						,(SETTINGS_PAGE[LOCAL_IP] & 0x00FF0000) >> 16 \
						,(SETTINGS_PAGE[LOCAL_IP] & 0x0000FF00) >> 8 \
						,(SETTINGS_PAGE[LOCAL_IP] & 0x000000FF) >> 0);
	
	IP4_ADDR(&my_net_mask,(SETTINGS_PAGE[NETMASK] & 0xFF000000) >> 24 \
						 ,(SETTINGS_PAGE[NETMASK] & 0x00FF0000) >> 16 \
						 ,(SETTINGS_PAGE[NETMASK] & 0x0000FF00) >> 8 \
						 ,(SETTINGS_PAGE[NETMASK] & 0x000000FF) >> 0);	
	IP4_ADDR(&my_gateway,emacGATEWAY_ADDR0,emacGATEWAY_ADDR1,emacGATEWAY_ADDR2,emacGATEWAY_ADDR3);	
	netif_add(&emac_if, &my_ip_addr, &my_net_mask, &my_gateway, NULL, ethernetif_init, tcpip_input);
			
	/* make it the default interface */
    netif_set_default(&emac_if);

	/* bring it up */
    netif_set_up(&emac_if);	
		
 	/* create a new UDP connection */
 	listen_pc = netconn_new( NETCONN_UDP );
 	
 	/* set up the IP address of the remote host. This is one is 192.168.2.7 */ 	 
	addr.addr = htonl(SETTINGS_PAGE[PC_IP]);
	
	/* connect the connection to the remote host */	
	netconn_connect(listen_pc, &addr, SETTINGS_PAGE[PC_PORT] );
	
	/* Make datagram_recved callback function and bind to the local port */		
	udp_recv( listen_pc->pcb.udp, datagram_recved, NULL );
	netconn_bind( listen_pc , NULL , SETTINGS_PAGE[LOCAL_PORT] );		
	
	/* create a new netbuf. This buffer is for referencing ethernet data. */
	buf = netbuf_new();	
	comm_open = 1;
	
	/* Loop forever */	
	for( ;; ) {		
		if( xSemaphoreTake( spi_buffer_semaphore, ( portTickType ) 0xFFFF ) == pdTRUE ) {
			if( comm_open > 1 ) {										
				/* Reference the request data into net_buf */							
				netbuf_ref( buf , (int *)filter() , DATA_FRAGMENT_LENGTH );
				//netbuf_ref( buf , (int *)mac_ptr , DATA_FRAGMENT_LENGTH );
				/* Then send the datagram */		
				netconn_send( listen_pc , buf );												
			}							
		} 
	}
}
void datagram_recved(void *arg, struct udp_pcb *pcb, struct pbuf *p, struct ip_addr *addr, u16_t port) 
{
	static int settings[SETTINGS_LENGTH];	
	
	/* Supress compiler warnings */
	( void ) arg;
	( void ) addr;
	( void ) port;	
	
	/* If we are entered config mode before then this should be a config command.
	 * Otherwise, process the datagram
	 */
	if( config_mode ) {
		configure_device( &settings[0] , pcb , p );		
	} else {		
		process_message( &settings[0] , pcb , p );											 				 				    									
	}	
	
	/* Don't forget to free buffer */		
	pbuf_free(p);	
}
inline void process_message( int *settings , struct udp_pcb *pcb, struct pbuf *p )
{
	char *mes = (char *)(p->payload);
	int i;
	/* Enter configuration mode if requested */
	if( strncasecmp( mes , "Enter Config" , 12) == 0 ) {
		ADS1258_start_stop(0);
		/* When we are in config mode config_mode is 1 */
		config_mode = 1;				
		send_message("Entering configuration mode...\0" , pcb);
		if(settings[0] != 0x10) {
			/* If this is the first time we enter config mode than initialize
			 * settings. Otherwise do not change it so that user can test settings
			 * before flashing settings.
			 */			 		
			for( i = 0; i < SETTINGS_LENGTH; i++ )
				settings[i] = SETTINGS_PAGE[i];
		} 				 				    									
	}
	/* Start sending ADC data */
	if( strncasecmp( mes , "Send" , 4) == 0 ) {
		comm_open = 2;				 				 				    									
	}
	/* Stop sending ADC data */
	if( strncasecmp( mes , "Stop sending" , 12) == 0 ) {
		comm_open = 1;				 				 				    									
	}							
	/* This function returns avaliable RAM in the FREERTOS heap */	  
	if( strncasecmp( mes , "Mem Test" , 8) == 0 ) {
		mes[0] = (char)vPortUsedMem() ;
		mes[1] = (char)(vPortUsedMem() >> 8) ;				
		udp_send(pcb, p);								
	}
}
inline void configure_device( int *settings , struct udp_pcb *pcb, struct pbuf *p )
{
	int i;
	char *mes = (char *)(p->payload);
					
	if( strncasecmp( mes , "Update firmware" , 15) == 0 ) {
		mes[0] = boot_from_rom() + 48 ;
		udp_send(pcb, p);										 				 				    									
	}	
	/* Write settings to the flash. */
	if( strncasecmp( mes , "Save config" , 11) == 0 ) {
		mes[0] = flash_write( settings , SETTINGS_LENGTH , SETTINGS_PAGE_NUMBER ) + 48 ;			
		/* Initialize adc sub-system to reflect changes */
		adc_system_init();	
		udp_send(pcb, p);
		//ads_load_defaults();	
	}
	/* Send settings to the PC */
	if( strncasecmp( mes , "Read all settings" , 17) == 0 ) {
		send_block( (int *)settings , SETTINGS_LENGTH , pcb );					
	}
	/* Receive settings from PC */
	if( strncasecmp( mes , "Write all settings" , 18) == 0 ) {
		/* Message length should be settings length in bytes plus 20 bytes */		
		if ( p->len >= 20 +  SETTINGS_LENGTH * 4 ) {				
			for( i = 0; i < SETTINGS_LENGTH ; i++ )
				settings[i] = ( mes[20 + i * 4] << 0 )  \
							+ ( mes[21 + i * 4] << 8 )  \
							+ ( mes[22 + i * 4] << 16 ) \
							+ ( mes[23 + i * 4] << 24 );							
			send_message("All settings changed\0" , pcb); 
		} else {
			send_message("All settings change fails!\0" , pcb);
		}					
	}
	/* This command corrects offsets of the ADC sub-system */
	if( strncasecmp( mes , "Correct Offsets" , 15) == 0 ) {
		if( offset_correct( &settings[0] ) == 0 ) 
			send_message("Offsets are corrected\0" , pcb);
		else
			send_message("Offset corrections failed\0" , pcb);						
	}
	/* Exit config and start ADC sub-system */
	if( strncasecmp( mes , "Exit Config" , 11) == 0 ) {
		ADS1258_start_stop(1);
		config_mode = 0;			
		send_message("Exiting configuration mode...\0" , pcb);						
	}
}
inline void send_message(char *mes , struct udp_pcb *pcb) 
{
	char *ans;
	struct pbuf *res;		
	 
	res = pbuf_alloc(PBUF_TRANSPORT, strlen(mes) , PBUF_RAM);
	
	if(res == NULL) 
		return;
	/* payload pointer should contain the actual data to send.  */ 	
	ans = (char *)res->payload;	
	strncpy( ans , mes , strlen(mes) );	
	udp_send(pcb, res);		
	pbuf_free(res);
}
inline void send_block(int *block , short length , struct udp_pcb *pcb) 
{
	struct pbuf *res;		
	int *ans; 
	res = pbuf_alloc(PBUF_TRANSPORT, sizeof(int)*length , PBUF_RAM);
	
	if(res == NULL) 
		return;
	
	/* payload pointer should contain the actual data to send.  */	
	ans = (int *)res->payload;
	/* length is not in bytes but in number of items. */	
	memcpy( ans , block , length * sizeof(int));	
	udp_send(pcb, res);		
	pbuf_free(res);
}
inline char offset_correct( int *settings )
{
	/* First of all we should configure ADS1258s in fix mode to read offset from
	 * unused channels.
	 */
	ads_load_for_offset();
	/* We are making at least 15 reading for ADS to become stable enough. */
	adc_data[0].offset_correction = 15;
	adc_data[1].offset_correction = 15;
	/* Switch to variable mode for SPIo since interrupts use variable mode */
	//spi_config_var_mode();
	/* Start conversion and then wait conversion end. When offsets are read 
	 * correctly, spi_buffer_semaphore will be released. 
	 */
	ADS1258_start_stop(1);
	if( xSemaphoreTake( spi_buffer_semaphore, ( portTickType ) 0xFFFF ) == pdTRUE ) {
		/* Semaphore is released meaning that we can save this offset as new value */
		ADS1258_start_stop(0);	
		settings[ADC0_OFFSET] = adc_data[0].offset & 0xFFFF;
		settings[ADC1_OFFSET] = adc_data[1].offset & 0xFFFF;
		/* Load ADS defaults for correct operation. */
		ads_load_defaults();					
		return 0;
	}			
	/* Waiting for semaphore timeout meaning we couldn't read the offsets */
	return 1;			
}
